#include <stdio.h>
#include <mpi.h>

static const int master_rank = 0;

static char*        base_filename = (char*)NULL;
static float        domain_size_x;
static float        domain_size_y;
static float        domain_size_z;
static float        time_interval;
static int          number_of_particles;
static int          number_of_timesteps;
static int          timesteps_between_outputs;
static unsigned int random_seed;

#include "nbody_particle_star.h"
#include "nbody_particle_star_mpi.c"

int main (int argc, char** argv)
{ /* main */
    Particle* particle_array = (Particle*)NULL;
    int       timestep;
    int       number_of_processes, my_rank;
    int       mpi_error_code;

    mpi_error_code = MPI_Init(&argc, &argv);
    mpi_error_code = MPI_Comm_size(MPI_COMM_WORLD, &number_of_processes);
    mpi_error_code = MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
    Particle_get_arguments(argc, argv);
    particle_array = Particle_array_construct(number_of_particles);
    Particle_array_initialize(particle_array, number_of_particles);
    Particle_array_output(base_filename,
        particle_array, number_of_particles, 0,
        number_of_processes, my_rank);
    for (timestep = 1; timestep <= number_of_timesteps; timestep++) {
        Particle_array_calculate_forces(particle_array, number_of_particles,
            number_of_processes, my_rank);
        Particle_array_move(particle_array, number_of_particles,
            time_interval);
        if ((timestep % timesteps_between_outputs) == 0) {
            Particle_array_output(base_filename,
                particle_array, number_of_particles, timestep,
                number_of_processes, my_rank);
        } /* if ((timestep % timesteps_between_outputs) == 0) */
    } /* for timestep */
    if ((number_of_timesteps % timesteps_between_outputs) != 0) {
        Particle_array_output(base_filename,
            particle_array, number_of_particles, number_of_timesteps,
            number_of_processes, my_rank);
    } /* if ((number_of_timesteps % timesteps_between_outputs) != 0) */
    particle_array =
        Particle_array_destruct(particle_array, number_of_particles);
    mpi_error_code = MPI_Finalize();
} /* main */

