
/* nbody_particle_star.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "nbody_particle_star.h"

#define default_domain_size_x               1.0e+18 /* m  */
#define default_domain_size_y               1.0e+18 /* m  */
#define default_domain_size_z               1.0e+18 /* m  */
#define default_mass_maximum                1.0e+18 /* kg */
#define default_time_interval               1.0e+18 /* s  */
#define default_number_of_particles         1000
#define default_number_of_timesteps         100
#define default_timesteps_between_outputs   default_number_of_timesteps
#define default_random_seed                 12345
#define default_base_filename               "nbody"

#define base_filename_argument             1
#define number_of_particles_argument       (base_filename_argument+1)
#define number_of_timesteps_argument       (number_of_particles_argument+1)
#define timesteps_between_outputs_argument (number_of_timesteps_argument+1)
#define domain_size_x_argument           (timesteps_between_outputs_argument+1)
#define domain_size_y_argument             (domain_size_x_argument+1)
#define domain_size_z_argument             (domain_size_y_argument+1)
#define mass_maximum_argument              (domain_size_z_argument+1)
#define time_interval_argument             (mass_maximum_argument+1)
#define random_seed_argument               (time_interval_argument+1)

static float mass_maximum;

/*
 * Get command line arguments.
 */
void Particle_get_arguments (int argc, char** argv)
{ /* Particle_get_arguments */
   /*
    * number_of_particles, number_of_timesteps, timesteps_between_outputs,
    * domain_size_x, domain_size_y, domain_size_z,
    * mass_maximum, time_interval,
    * random_seed
    */
    number_of_particles       = default_number_of_particles;
    domain_size_x             = default_domain_size_x;
    domain_size_y             = default_domain_size_y;
    domain_size_z             = default_domain_size_z;
    time_interval             = default_time_interval;
    number_of_timesteps       = default_number_of_timesteps;
    timesteps_between_outputs = default_timesteps_between_outputs;
    random_seed               = default_random_seed;
    mass_maximum              = default_mass_maximum;
    if (argc > base_filename_argument) {
        if (strlen(argv[base_filename_argument]) < 1) {
            base_filename = strdup(default_base_filename);
        } /* if (strlen(argv[base_filename_argument]) < 1) */
        else {
            base_filename = strdup(argv[base_filename_argument]);
        } /* if (strlen(argv[base_filename_argument]) < 1) */
    } /* if (argc > base_filename_argument) */
    else {
        base_filename = strdup(default_base_filename);
    } /* if (argc > base_filename_argument)...else */
    if (argc > number_of_particles_argument) {
        if (sscanf(argv[number_of_particles_argument], "%d",
                  &number_of_particles) != 1) {
            fprintf(stderr, "ERROR: illegal number of particles\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                number_of_particles_argument,
                argv[number_of_particles_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%d", &number_of_particles) != 1) */
    } /* if (argc > number_of_particles_argument) */
    if (number_of_particles < 1) {
        fprintf(stderr, "ERROR: bad number of particles %d\n",
            number_of_particles);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (number_of_particles < 1) */
    if (argc > number_of_timesteps_argument) {
        if (sscanf(argv[number_of_timesteps_argument], "%d",
                  &number_of_timesteps) != 1) {
            fprintf(stderr, "ERROR: illegal number of timesteps\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                number_of_timesteps_argument,
                argv[number_of_timesteps_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%d", &number_of_timesteps) != 1) */
    } /* if (argc > number_of_timesteps_argument) */
    if (number_of_timesteps < 1) {
        fprintf(stderr, "ERROR: bad number of timesteps %d\n",
            number_of_timesteps);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (number_of_timesteps < 1) */
    if (argc > timesteps_between_outputs_argument) {
        if (sscanf(argv[timesteps_between_outputs_argument], "%d",
                  &timesteps_between_outputs) != 1) {
            fprintf(stderr,
                "ERROR: illegal number of timesteps between outputs\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                timesteps_between_outputs_argument,
                argv[timesteps_between_outputs_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%d", &timesteps_between_outputs) != 1) */
    } /* if (argc > timesteps_between_outputs_argument) */
    if (timesteps_between_outputs < 1) {
        fprintf(stderr, "ERROR: bad number of timesteps between outputs %d\n",
            timesteps_between_outputs);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (timesteps_between_outputs < 1) */
    if (argc > domain_size_x_argument) {
        if (sscanf(argv[domain_size_x_argument], "%g",
                  &domain_size_x) != 1) {
            fprintf(stderr, "ERROR: illegal domain size along the X axis\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                domain_size_x_argument,
                argv[domain_size_x_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%g", &domain_size_x) != 1) */
    } /* if (argc > domain_size_x_argument) */
    if (domain_size_x <= 0.0) {
        fprintf(stderr, "ERROR: bad domain size along the X axis %g\n",
            domain_size_x);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (domain_size_x <= 0.0) */
    if (argc > domain_size_y_argument) {
        if (sscanf(argv[domain_size_y_argument], "%g",
                  &domain_size_y) != 1) {
            fprintf(stderr, "ERROR: illegal domain size along the Y axis\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                domain_size_y_argument,
                argv[domain_size_y_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%g", &domain_size_y) != 1) */
    } /* if (argc > domain_size_y_argument) */
    if (domain_size_y <= 0.0) {
        fprintf(stderr, "ERROR: bad domain size along the Y axis %g\n",
            domain_size_y);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (domain_size_y <= 0.0) */
    if (argc > domain_size_z_argument) {
        if (sscanf(argv[domain_size_z_argument], "%g",
                  &domain_size_z) != 1) {
            fprintf(stderr, "ERROR: illegal domain size along the Z axis\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                domain_size_z_argument,
                argv[domain_size_z_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%g", &domain_size_z) != 1) */
    } /* if (argc > domain_size_z_argument) */
    if (domain_size_z <= 0.0) {
        fprintf(stderr, "ERROR: bad domain size along the Z axis %g\n",
            domain_size_z);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (domain_size_z <= 0.0) */
    if (argc > mass_maximum_argument) {
        if (sscanf(argv[mass_maximum_argument], "%g",
                  &mass_maximum) != 1) {
            fprintf(stderr, "ERROR: illegal mass maximum\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                mass_maximum_argument,
                argv[mass_maximum_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%g", &mass_maximum) != 1) */
    } /* if (argc > mass_maximum_argument) */
    if (mass_maximum <= 0.0) {
        fprintf(stderr, "ERROR: bad mass maximum %g\n",
            mass_maximum);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (mass_maximum <= 0.0) */
    if (argc > time_interval_argument) {
        if (sscanf(argv[time_interval_argument], "%g",
                  &time_interval) != 1) {
            fprintf(stderr, "ERROR: illegal time interval\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                time_interval_argument,
                argv[time_interval_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%g", &time_interval) != 1) */
    } /* if (argc > time_interval_argument) */
    if (time_interval <= 0.0) {
        fprintf(stderr, "ERROR: bad time interval %g\n",
            time_interval);
        fprintf(stderr, "  in Particle_get_arguments\n");
        exit(-1);
    } /* if (time_interval <= 0.0) */
    if (argc > random_seed_argument) {
        if (sscanf(argv[random_seed_argument], "%d",
                  &random_seed) != 1) {
            fprintf(stderr, "ERROR: illegal random seed\n");
            fprintf(stderr, "  in command line argument #%d:  %s\n",
                random_seed_argument,
                argv[random_seed_argument]);
            fprintf(stderr, "  in Particle_get_arguments\n");
            exit(-1);
        } /* if (sscanf(argv[...], "%d", &random_seed) != 1) */
    } /* if (argc > random_seed_argument) */
} /* Particle_get_arguments */

#ifdef CHECK
/*
 * Check that the particle exists.
 */
void Particle_check (Particle* this_particle, char* action, char* routine)
{ /* Particle_check */
    if (this_particle != (Particle*)NULL) return;
    fprintf(stderr, "ERROR: can't %s a nonexistent particle\n",
        ((action == (char*)NULL) || (strlen(action) == 0)) ?
            "perform an unknown action on" : action);
    fprintf(stderr, "  in %s\n",
        ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
            "an unknown routine" : routine);
    exit(-1);
} /* Particle_check */

/*
 * Check that the pair of particles both exist.
 */
void Particle_pair_check (Particle* this_particle1, Particle* this_particle2,
         char* action, char* routine)
{ /* Particle_pair_check */
    if ((this_particle1 != (Particle*)NULL) &&
        (this_particle2 != (Particle*)NULL))
        return;
    fprintf(stderr, "ERROR: can't %s\n",
        ((action == (char*)NULL) || (strlen(action) == 0)) ?
            "perform an unknown action on" : action);
    if (this_particle1 == (Particle*)NULL) {
        if (this_particle2 == (Particle*)NULL) {
            fprintf(stderr, "  a pair of nonexistent particles\n");
        } /* if (this_particle2 == (Particle*)NULL) */
        else {
            fprintf(stderr,
                "  a nonexistent particle and an existing particle\n");
        } /* if (this_particle2 == (Particle*)NULL)...else */
    } /* if (this_particle1 == (Particle*)NULL) */
    else {
        fprintf(stderr,
            "  an existing particle and a nonexistent particle\n");
    } /* if (this_particle1 == (Particle*)NULL)...else */
    fprintf(stderr, "  in %s\n",
        ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
            "an unknown routine" : routine);
    exit(-1);
} /* Particle_pair_check */
#endif /* #ifdef CHECK */

/*
 * Clear the particle's data.
 */
void Particle_clear (Particle* this_particle)
{ /* Particle_clear */
#ifdef CHECK
    Particle_check(this_particle, "clear", "Particle_clear");
#endif /* #ifdef CHECK */
    this_particle->position_x = 0.0;
    this_particle->position_y = 0.0;
    this_particle->position_z = 0.0;
    this_particle->velocity_x = 0.0;
    this_particle->velocity_y = 0.0;
    this_particle->velocity_z = 0.0;
    this_particle->force_x    = 0.0;
    this_particle->force_y    = 0.0;
    this_particle->force_z    = 0.0;
    this_particle->mass       = 0.0;
} /* Particle_clear */

/*
 * Construct the particle.
 */
void Particle_construct (Particle* this_particle)
{ /* Particle_construct */
#ifdef CHECK
    Particle_check(this_particle, "construct", "Particle_construct");
#endif /* #ifdef CHECK */
    Particle_clear(this_particle);
} /* Particle_construct */

/*
 * Destroy the particle.
 */
void Particle_destruct (Particle* this_particle)
{ /* Particle_destruct */
#ifdef CHECK
    Particle_check(this_particle, "destruct", "Particle_destruct");
#endif /* #ifdef CHECK */
    Particle_clear(this_particle);
} /* Particle_destruct */

/*
 * Initialize the particle by setting its data randomly.
 */
void Particle_set_position_randomly (Particle* this_particle)
{ /* Particle_set_position_randomly */
#ifdef CHECK
    Particle_check(this_particle, "randomly set the position",
        "Particle_set_randomly");
#endif /* #ifdef CHECK */
    this_particle->position_x =
        domain_size_x * (float)(random() / ((float)RAND_MAX + 1.0));
    this_particle->position_y =
        domain_size_y * (float)(random() / ((float)RAND_MAX + 1.0));
    this_particle->position_z =
        domain_size_z * (float)(random() / ((float)RAND_MAX + 1.0));
} /* Particle_set_position_randomly */

/*
 * Initialize the particle by setting its data randomly.
 */
void Particle_initialize_randomly (Particle* this_particle)
{ /* Particle_initialize_randomly */
#ifdef CHECK
    Particle_check(this_particle, "randomly initialize",
        "Particle_initialize_randomly");
#endif /* #ifdef CHECK */
    Particle_clear(this_particle);
    Particle_set_position_randomly(this_particle);
    this_particle->mass       =
        mass_maximum  * (float)(random() / ((float)RAND_MAX + 1.0));
} /* Particle_initialize_randomly */

/*
 * Initialize the particle.
 */
void Particle_initialize (Particle* this_particle)
{ /* Particle_initialize */
#ifdef CHECK
    Particle_check(this_particle, "initialize", "Particle_initialize");
#endif /* #ifdef CHECK */
    Particle_initialize_randomly(this_particle);
} /* Particle_initialize */

/*
 * Calculate the force between the two particles.
 *
 * Gregory R. Andrews, "Foundations of Multithreaded, Parallel, and
 * Distributed Programming," Addison-Wesley, 2000, p. 554-6.
 */
void Particle_calculate_force (
         Particle* this_particle1, Particle* this_particle2,
         float* force_x, float* force_y, float* force_z)
{ /* Particle_calculate_force */
    float difference_x, difference_y, difference_z;
    float distance_squared, distance;
    float force_magnitude;

#ifdef CHECK
    Particle_pair_check(this_particle1, this_particle2,
        "calculate the force along the X-axis", "Particle_calculate_force_x");
#endif /* #ifdef CHECK */
    difference_x =
        this_particle2->position_x - this_particle1->position_x;
    difference_y =
        this_particle2->position_y - this_particle1->position_y;
    difference_z =
        this_particle2->position_z - this_particle1->position_z;
    distance_squared = difference_x * difference_x +
                       difference_y * difference_y +
                       difference_z * difference_z;
    distance = sqrt(distance_squared);
    force_magnitude =
        gravitational_constant *
        (this_particle1->mass) * (this_particle2->mass) / distance_squared;
    *force_x = (force_magnitude / distance) * difference_x;
    *force_y = (force_magnitude / distance) * difference_y;
    *force_z = (force_magnitude / distance) * difference_z;
} /* Particle_calculate_force */

/*
 * Move the particle based on the force applied to it.
 *
 * Gregory R. Andrews, "Foundations of Multithreaded, Parallel, and
 * Distributed Programming," Addison-Wesley, 2000, p. 554-6.
 */
void Particle_move (Particle* this_particle, float time_interval)
{ /* Particle_move */
    float velocity_change_x, velocity_change_y, velocity_change_z;
    float position_change_x, position_change_y, position_change_z;

#ifdef CHECK
    Particle_check(this_particle, "move", "Particle_move");
#endif /* #ifdef CHECK */
    velocity_change_x =
        this_particle->force_x * (time_interval / this_particle->mass);
    velocity_change_y =
        this_particle->force_y * (time_interval / this_particle->mass);
    velocity_change_z =
        this_particle->force_z * (time_interval / this_particle->mass);
    position_change_x =
        this_particle->velocity_x + velocity_change_x * (0.5 * time_interval);
    position_change_y =
        this_particle->velocity_y + velocity_change_y * (0.5 * time_interval);
    position_change_z =
        this_particle->velocity_z + velocity_change_z * (0.5 * time_interval);
    this_particle->velocity_x += velocity_change_x;
    this_particle->velocity_y += velocity_change_y;
    this_particle->velocity_z += velocity_change_z;
    this_particle->position_x += position_change_x;
    this_particle->position_y += position_change_y;
    this_particle->position_z += position_change_z;
} /* Particle_move */

void Particle_output (FILE* fileptr, Particle* this_particle)
{ /* Particle_output */
    fprintf(fileptr, "%g %g %g %g %g %g %g\n",
        this_particle->position_x,
        this_particle->position_y,
        this_particle->position_z,
        this_particle->velocity_x,
        this_particle->velocity_y,
        this_particle->velocity_z,
        this_particle->mass);
} /* Particle_output */

#ifdef CHECK
void Particle_array_check (
         Particle* this_particle_array, int number_of_particles,
         char* action, char* routine)
{ /* Particle_array_check */
    if (number_of_particles < 0) {
        fprintf(stderr, "ERROR: illegal number of particles %d\n",
            number_of_particles);
        fprintf(stderr, "  to %s\n", 
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);
        exit(-1);
    } /* if (number_of_particles < 0) */
    if (number_of_particles == 0) {
        if (this_particle_array == (Particle*)NULL) return (Particle*)NULL;
        fprintf(stderr, "ERROR: can't %s\n",
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  an existing particle array of length 0\n");
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);
        exit(-1);
    } /* if (number_of_particles == 0) */
    if (this_particle_array == (Particle*)NULL) {
        fprintf(stderr, "ERROR: can't %s\n",
            ((action == (char*)NULL) || (strlen(action) == 0)) ?
                "perform an unknown action on" : action);
        fprintf(stderr, "  a nonexistent array of %d particles\n",
            number_of_particles);
        fprintf(stderr, "  in %s\n",
            ((routine == (char*)NULL) || (strlen(routine) == 0)) ?
                "an unknown routine" : routine);
        exit(-1);
    } /* if (this_particle_array == (Particle*)NULL) */
} /* Particle_array_check */
#endif /* #ifdef CHECK */

/*
 * Allocate and return an array of particles.
 */
Particle* Particle_array_allocate (int number_of_particles)
{ /* Particle_array_allocate */
    Particle* this_particle_array = (Particle*)NULL;
    int index;

#ifdef CHECK
    if (number_of_particles < 0) {
        fprintf(stderr,
            "ERROR: illegal number of particles %d to allocate\n", 
            number_of_particles);
        fprintf(stderr, "  in Particle_array_construct\n");
        exit(-1);
    } /* if (number_of_particles < 0) */
#endif /* #ifdef CHECK */
    if (number_of_particles == 0) return (Particle*)NULL;
    this_particle_array =
        (Particle*)malloc(sizeof(Particle) * number_of_particles);
    if (this_particle_array == (Particle*)NULL) {
        fprintf(stderr,
            "ERROR: can't allocate a particle array of %d particles\n", 
            number_of_particles);
        fprintf(stderr, "  in Particle_array_construct\n");
        exit(-1);
    } /* if (this_particle_array == (Particle*)NULL) */
    return this_particle_array;
} /* Particle_array_allocate */

/*
 * Construct and return an array of particles, cleared.
 */
Particle* Particle_array_construct (int number_of_particles)
{ /* Particle_array_construct */
    Particle* this_particle_array = (Particle*)NULL;
    int index;

    this_particle_array =
        Particle_array_allocate(number_of_particles);
    for (index = 0; index < number_of_particles; index++) {
        Particle_construct(&(this_particle_array[index]));
    } /* for index */
    return this_particle_array;
} /* Particle_array_construct */

/*
 * Deallocate the array of particles, and return NULL.
 */
Particle* Particle_array_deallocate (Particle* this_particle_array,
              int number_of_particles)
{ /* Particle_array_deallocate */
    int index;

#ifdef CHECK
    Particle_array_check(this_particle_array, number_of_particles,
        "deallocate", "Particle_array_deallocate");
#endif /* #ifdef CHECK */
    free(this_particle_array);
    this_particle_array = (Particle*)NULL;
    return (Particle*)NULL;
} /* Particle_array_deallocate */

/*
 * Destroy the array of particles, and return NULL.
 */
Particle* Particle_array_destruct (Particle* this_particle_array,
              int number_of_particles)
{ /* Particle_array_destruct */
    int index;

#ifdef CHECK
    Particle_array_check(this_particle_array, number_of_particles,
        "destroy", "Particle_array_destruct");
#endif /* #ifdef CHECK */
    for (index = number_of_particles - 1; index >= 0; index--) {
        Particle_destruct(&(this_particle_array[index]));
    } /* for index */
    return Particle_array_deallocate(this_particle_array, number_of_particles);
} /* Particle_array_destruct */

/*
 * Initialize the array of particles by setting its data randomly.
 */
void Particle_array_initialize_randomly (
         Particle* this_particle_array, int number_of_particles)
{ /* Particle_array_initialize_randomly */
    int index;

#ifdef CHECK
    Particle_array_check(this_particle_array, number_of_particles,
        "initialize randomly", "Particle_array_initialize_randomly");
#endif /* #ifdef CHECK */
    for (index = 0; index < number_of_particles; index++) {
        Particle_initialize_randomly(&(this_particle_array[index]));
    } /* for index */
} /* Particle_array_initialize_randomly */

/*
 * Initialize the array of particles.
 */
void Particle_array_initialize (
         Particle* this_particle_array, int number_of_particles)
{ /* Particle_array_initialize */
    Particle_array_initialize_randomly(
        this_particle_array, number_of_particles);
} /* Particle_array_initialize */

/*
 * Calculate the forces between every pair of particles, computing
 * every pair twice to keep the loops simple.
 */
void Particle_array_calculate_forces_simply (
         Particle* this_particle_array, int number_of_particles)
{ /* Particle_array_calculate_forces_simply */
    float particle_pair_force_x, particle_pair_force_y, particle_pair_force_z;
    int   index1, index2;

#ifdef CHECK
    Particle_array_check(this_particle_array, number_of_particles,
        "calculate forces simply", "Particle_array_calculate_forces_simply");
#endif /* #ifdef CHECK */
    if (number_of_particles <= 1) return;
    for (index1 = 0; index1 < number_of_particles; index1++) {
        this_particle_array[index1].force_x = 0.0;
        this_particle_array[index1].force_y = 0.0;
        this_particle_array[index1].force_z = 0.0;
    } /* for index1 */
    for (index1 = 0; index1 < number_of_particles; index1++) {
        for (index2 = 0; index2 < number_of_particles; index2++) {
            if (index1 != index2) {
                Particle_calculate_force(
                    &(this_particle_array[index1]),
                    &(this_particle_array[index2]),
                    &particle_pair_force_x,
                    &particle_pair_force_y,
                    &particle_pair_force_z);
                this_particle_array[index1].force_x += particle_pair_force_x;
                this_particle_array[index1].force_y += particle_pair_force_y;
                this_particle_array[index1].force_z += particle_pair_force_z;
            } /* if (index1 != index2) */
        } /* for index2 */
    } /* for index1 */
} /* Particle_array_calculate_forces_simply */

/*
 * Calculate the forces between every pair of particles, computing
 * each pair only once.
 */
void Particle_array_calculate_forces_quickly (
         Particle* this_particle_array, int number_of_particles)
{ /* Particle_array_calculate_forces_quickly */
    float particle_pair_force_x, particle_pair_force_y, particle_pair_force_z;
    int   index1, index2;

#ifdef CHECK
    Particle_array_check(this_particle_array, number_of_particles,
        "calculate forces quickly", "Particle_array_calculate_forces_quickly");
#endif /* #ifdef CHECK */
    if (number_of_particles <= 1) return;
    for (index1 = 0; index1 < number_of_particles; index1++) {
        this_particle_array[index1].force_x = 0.0;
        this_particle_array[index1].force_y = 0.0;
        this_particle_array[index1].force_z = 0.0;
    } /* for index1 */
    for (index1 = 1; index1 < number_of_particles; index1++) {
        for (index2 = 0; index2 < index1; index2++) {
            Particle_calculate_force(
                &(this_particle_array[index1]),
                &(this_particle_array[index2]),
                &particle_pair_force_x,
                &particle_pair_force_y,
                &particle_pair_force_z);
            this_particle_array[index1].force_x += particle_pair_force_x;
            this_particle_array[index1].force_y += particle_pair_force_y;
            this_particle_array[index1].force_z += particle_pair_force_z;
            this_particle_array[index2].force_x -= particle_pair_force_x;
            this_particle_array[index2].force_y -= particle_pair_force_y;
            this_particle_array[index2].force_z -= particle_pair_force_z;
        } /* for index2 */
    } /* for index1 */
} /* Particle_array_calculate_forces_quickly */

/*
 * Calculate the forces between every pair of particles.
 */
void Particle_array_calculate_forces (
         Particle* this_particle_array, int number_of_particles)
{ /* Particle_array_calculate_forces */
    Particle_array_calculate_forces_quickly(
        this_particle_array, number_of_particles);
} /* Particle_array_calculate_forces */

/*
 * Move every particle in the array of particles.
 */
void Particle_array_move (
         Particle* this_particle_array, int number_of_particles,
         float time_interval)
{ /* Particle_array_move */
    int index;

#ifdef CHECK
    Particle_array_check(this_particle_array, number_of_particles,
        "calculate forces quickly", "Particle_array_calculate_forces_quickly");
#endif /* #ifdef CHECK */
    if (number_of_particles < 1) return;
    for (index = 0; index < number_of_particles; index++) {
        Particle_move(&(this_particle_array[index]), time_interval);
    } /* for index */
} /* Particle_array_move */

void Particle_array_output (char* base_filename,
         Particle* this_particle_array, int number_of_particles, int timestep)
{ /* Particle_array_output */
    FILE* fileptr  = (FILE*)NULL;
    char* filename = (char*)NULL;
    int   filename_length;
    int   index;

#ifdef CHECK
    Particle_array_check(this_particle_array, number_of_particles,
        "output", "Particle_array_output");
#endif /* #ifdef CHECK */
    filename_length = strlen(base_filename) + 1 + 8 + 1 + 3;
    filename = (char*)malloc(sizeof(char) * (filename_length + 1));
    if (filename == (char*)NULL) {
        fprintf(stderr, "ERROR: can't allocate the filename string\n");
        fprintf(stderr, "  %s_%8.8d.txt\n", base_filename, timestep);
        fprintf(stderr, "  in Particle_array_output\n");
        exit(-1);
    } /* if (filename == (char*)NULL) */
    sprintf(filename, "%s_%8.8d.txt", base_filename, timestep);
    fileptr = fopen(filename, "w");
    if (fileptr == (FILE*)NULL) {
        fprintf(stderr, "ERROR: can't open the output file named\n");
        fprintf(stderr, "  %s\n", filename);
        fprintf(stderr, "  in Particle_array_output\n");
        exit(-1);
    } /* if (fileptr == (FILE*)NULL) */
    fprintf(fileptr, "%d %d %d %g %g %g %g %g %d\n",
        number_of_particles, number_of_timesteps, timesteps_between_outputs,
        domain_size_x, domain_size_y, domain_size_z,
        mass_maximum, time_interval,
        random_seed);
    fprintf(fileptr, "%d\n", timestep);
    for (index = 0; index < number_of_particles; index++) {
        Particle_output(fileptr, &(this_particle_array[index]));
    } /* for index */
    if (fclose(fileptr) != 0) {
        fprintf(stderr, "ERROR: can't close the output file named\n");
        fprintf(stderr, "  %s\n", filename);
        fprintf(stderr, "  in Particle_array_output\n");
        exit(-1);
    } /* if (fclose(fileptr) != 0) */
    fileptr = (FILE*)NULL;
    free(filename); filename = (char*)NULL;
} /* Particle_array_output */

